#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009  Douglas S. Blank <doug.blank@gmail.com>
# Copyright (C) 2010  Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Base class for Gramplet plugins
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import types
import logging

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from ...gui.dbguielement import DbGUIElement
from ..const import GRAMPS_LOCALE as glocale

LOG = logging.getLogger(".Gramplets")

_ = glocale.translation.gettext


# ------------------------------------------------------------------------
#
# Gramplet class
#
# ------------------------------------------------------------------------
class Gramplet:
    """
    Base class for non-graphical gramplet code.
    """

    def __init__(self, gui, nav_group=0):
        """
        Internal constructor for non-graphical gramplets.
        """
        self._idle_id = 0
        self.track = []
        self.active = False
        self.dirty = True
        self.has_data = True
        self._pause = False
        self._generator = None
        self._need_to_update = False
        self.option_dict = {}
        self._signal = {}
        self.option_order = []
        # links to each other:
        self.gui = gui  # plugin gramplet has link to gui
        gui.pui = self  # gui has link to plugin ui
        self.nav_group = nav_group
        self.dbstate = gui.dbstate
        self.uistate = gui.uistate
        self.init()
        self.on_load()
        self.build_options()
        self.dbstate.connect("database-changed", self._db_changed)
        self.dbstate.connect("no-database", self._no_db)
        self.gui.textview.connect("button-press-event", self.gui.on_button_press)
        self.gui.textview.connect("motion-notify-event", self.gui.on_motion)
        self._db_changed(self.dbstate.db)
        active_person = self.get_active("Person")
        if active_person:  # already changed
            self._active_changed(active_person)
        self.post_init()

    def connect_signal(self, nav_type, method):
        """
        Connect the given method to the active-changed signal for the
        navigation type requested.
        """
        self.uistate.register(self.dbstate, nav_type, self.nav_group)
        history = self.uistate.get_history(nav_type, self.nav_group)
        # print('History: nave-type = %s' % nav_type)
        self.connect(history, "active-changed", method)

    def init(self):  # once, constructor
        """
        External constructor for developers to put their initialization
        code. Designed to be overridden.
        """

    def post_init(self):
        """
        Another for developers to hook into post initialization phase.
        """

    def build_options(self):
        """
        External constructor for developers to put code for building
        options.
        """

    def main(self):  # return false finishes
        """
        The main place for the gramplet's code. This is a generator.
        Generator which will be run in the background, through :meth:`update`.
        """
        yield False

    def on_load(self):
        """
        Gramplets should override this to take care of loading previously
        their special data.
        """

    def on_save(self):
        """
        Gramplets should override this to take care of saving their
        special data.
        """

    def get_active(self, nav_type):
        """
        Return the handle of the active object for the given navigation type.
        """
        return self.uistate.get_active(nav_type, self.nav_group)

    def get_active_object(self, nav_type):
        """
        Return the object of the active handle for the given navigation type.
        """
        handle = self.uistate.get_active(nav_type, self.nav_group)
        handle_func = getattr(self.dbstate.db, f"get_{nav_type.lower()}_from_handle")
        if handle:
            return handle_func(handle)
        return None

    def set_active(self, nav_type, handle):
        """
        Change the handle of the active object for the given navigation type.
        """
        self.uistate.set_active(handle, nav_type, self.nav_group)

    def active_changed(self, handle):
        """
        Developers should put their code that occurs when the active
        person is changed.
        """

    def _active_changed(self, handle):
        """
        Private code that updates the GUI when active_person is changed.
        """
        self.active_changed(handle)

    def db_changed(self):
        """
        Method executed when the database is changed.
        """

    def link(self, text, link_type, data, size=None, tooltip=None):
        """
        Creates a clickable link in the textview area.
        """
        self.gui.link(text, link_type, data, size, tooltip)

    # Shortcuts to the gui functionality:

    def set_tooltip(self, tip):
        """
        Sets the tooltip for this gramplet.
        """
        self.gui.set_tooltip(tip)

    def get_text(self):
        """
        Returns the current text of the textview.
        """
        return self.gui.get_text()

    def insert_text(self, text):
        """
        Insert the given text in the textview at the cursor.
        """
        self.gui.insert_text(text)

    def render_text(self, text):
        """
        Render the given text, given that :meth:`set_use_markup` is on.
        """
        self.gui.render_text(text)

    def clear_text(self):
        """
        Clear all of the text from the textview.
        """
        self.gui.clear_text()

    def set_text(self, text, scroll_to="start"):
        """
        Clear and set the text to the given text. Additionally, move the
        cursor to the position given. Positions are:

            ========  =======================================
            Position  Description
            ========  =======================================
            'start'   start of textview
            'end'     end of textview
            'begin'   begin of line, before setting the text.
            ========  =======================================
        """
        self.gui.set_text(text, scroll_to)

    def append_text(self, text, scroll_to="end"):
        """
        Append the text to the textview. Additionally, move the
        cursor to the position given. Positions are:

            ========  =======================================
            Position  Description
            ========  =======================================
            'start'   start of textview
            'end'     end of textview
            'begin'   begin of line, before setting the text.
            ========  =======================================
        """
        self.gui.append_text(text, scroll_to)

    def set_use_markup(self, value):
        """
        Allows the use of render_text to show markup.
        """
        self.gui.set_use_markup(value)

    def set_wrap(self, value):
        """
        Set the textview to wrap or not.
        """
        textview = self.gui.textview
        from gi.repository import Gtk

        # Gtk.WrapMode.NONE, Gtk.WrapMode.CHAR, Gtk.WrapMode.WORD or Gtk.WrapMode.WORD_CHAR.
        if value in [True, 1]:
            textview.set_wrap_mode(Gtk.WrapMode.WORD)
        elif value in [False, 0, None]:
            textview.set_wrap_mode(Gtk.WrapMode.NONE)
        elif value in ["char"]:
            textview.set_wrap_mode(Gtk.WrapMode.CHAR)
        elif value in ["word char"]:
            textview.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        else:
            raise ValueError(
                f"Unknown wrap mode: '{value}': use 0,1,'char' or 'word char')"
            )

    def no_wrap(self):
        """
        The view in gramplet should not wrap.
        DEPRICATED: use :meth:`set_wrap` instead.
        """
        self.set_wrap(False)

    # Other functions of the gramplet:

    def load_data_to_text(self, pos=0):
        """
        Load information from the data portion of the saved
        Gramplet to the textview.
        """
        if len(self.gui.data) >= pos + 1:
            text = self.gui.data[pos]
            text = text.replace("\\n", chr(10))
            self.set_text(text, "end")

    def save_text_to_data(self):
        """
        Save the textview to the data portion of a saved gramplet.
        """
        text = self.get_text()
        text = text.replace(chr(10), "\\n")
        self.gui.data.append(text)

    def update(self, *_):
        """
        The main interface for running the :meth:`main` method.
        """
        from gi.repository import GLib

        if (not self.active) and not self.gui.force_update:
            self.dirty = True
            if self.dbstate.is_open():
                # print "  %s is not active" % self.gui.gname
                self.update_has_data()
            else:
                self.set_has_data(False)
            return
        # print "     %s is UPDATING" % self.gui.gname
        self.dirty = False
        LOG.debug("gramplet updater: %s: running", self.gui.title)
        if self._idle_id != 0:
            self.interrupt()
        self._generator = self.main()
        self._pause = False
        self._idle_id = GLib.idle_add(self._updater, priority=GLib.PRIORITY_LOW - 10)

    def _updater(self):
        """
        Runs the generator.
        """
        LOG.debug("gramplet updater: %s", self.gui.title)
        if not isinstance(self._generator, types.GeneratorType):
            self._idle_id = 0
            LOG.debug("gramplet updater: %s : One time, done!", self.gui.title)
            return False
        try:
            retval = next(self._generator)
            if not retval:
                self._idle_id = 0
            if self._pause:
                LOG.debug("gramplet updater: %s: return False", self.gui.title)
                return False
            LOG.debug("gramplet updater: %s: return %s", self.gui.title, retval)
            return retval
        except StopIteration:
            self._idle_id = 0
            self._generator.close()
            LOG.debug("gramplet updater: %s: Done!", self.gui.title)
            return False
        except Exception:
            import traceback

            LOG.warning("Gramplet gave an error: %s", self.gui.title)
            traceback.print_exc()
            print("Continuing after gramplet error...")
            self._idle_id = 0
            self.uistate.push_message(
                self.dbstate, _("Gramplet %s caused an error") % self.gui.title
            )
            return False

    def pause(self, *_):
        """
        Pause the :meth:`main` method.
        """
        self._pause = True

    def resume(self, *_):
        """
        Resume the :meth:`main` method that has previously paused.
        """
        from gi.repository import GLib

        self._pause = False
        self._idle_id = GLib.idle_add(self._updater, priority=GLib.PRIORITY_LOW - 10)

    def update_all(self, *_):
        """
        Force the main loop to run right now (as opposed to running in
        background).
        """
        self._generator = self.main()
        if isinstance(self._generator, types.GeneratorType):
            for _ in self._generator:
                pass

    def interrupt(self, *_):
        """
        Force the generator to stop running.
        """
        from gi.repository import GLib

        self._pause = True
        if self._idle_id != 0:
            GLib.source_remove(self._idle_id)
            self._idle_id = 0

    def _db_changed(self, db):
        """
        Internal method for handling items that should happen when the
        database changes. This will push a message to the GUI status bar.
        """
        if self.dbstate.db.is_open():
            self.disconnect_all()  # clear the old signals from old db
        self.dbstate.db = db
        self.gui.dbstate.db = db
        if db.is_open():
            # the following prevents connecting to every Gramplet, and still
            # allows Person Gramplets to be informed of active-changed
            # Some Gramplets .gpr files don't have navtypes set, thus the
            # hasattr
            if hasattr(self.gui, "navtypes") and "Person" in self.gui.navtypes:
                self.connect_signal("Person", self._active_changed)
            self.db_changed()
        # Some Gramplets use DbGUIElement; and DbGUIElement needs to know if
        # db is changed.  However, at initialization, DbGUIElement is not yet
        # initialized when _db_changed is called, thus the test for callman
        if hasattr(self, "callman") and isinstance(self, DbGUIElement):
            # get DbGUIElement informed if in use
            self._change_db(db)  # DbGUIElement method

        self.update()

    def _no_db(self):
        self.disconnect_all()  # clear the old signals

    def get_option_widget(self, label):
        """
        Retrieve an option's widget by its label text.
        """
        return self.option_dict[label][0]

    def get_option(self, label):
        """
        Retrieve an option by its label text.
        """
        return self.option_dict[label][1]

    def add_option(self, option):
        """
        Add an option to the GUI gramplet.
        """
        widget, _ = self.gui.add_gui_option(option)
        self.option_dict.update({option.get_label(): [widget, option]})
        self.option_order.append(option.get_label())

    def save_update_options(self, obj):
        """
        Save a gramplet's options to file.
        """
        self.save_options()
        self.update()

    def save_options(self):
        """
        Save a gramplet's options.
        """

    def connect(self, signal_obj, signal, method):
        """
        Connect signals.
        """
        signal_id = signal_obj.connect(signal, method)
        signal_list = self._signal.get(signal, [])
        signal_list.append((signal_id, signal_obj))
        self._signal[signal] = signal_list

    def disconnect(self, signal):
        """
        Disconnect signals.
        """
        if signal in self._signal:
            for signal_id, signal_obj in self._signal[signal]:
                signal_obj.disconnect(signal_id)
            self._signal[signal] = []
        else:
            raise AttributeError(f"unknown signal: '{signal}'")

    def disconnect_all(self):
        """
        Used to disconnect all the signals for this specific gramplet
        """
        for signal, signal_data in self._signal.items():
            for signal_id, signal_obj in signal_data:
                signal_obj.disconnect(signal_id)
            self._signal[signal] = []

    def hidden_widgets(self):
        """
        A list of widgets to keep hidden. Needed because Gramps uses
        show_all() in some places.
        """
        return []

    def set_has_data(self, value):
        """
        Set the status as to whether this gramplet has data.
        """
        if value != self.has_data:
            self.has_data = value
            self.gui.set_has_data(value)

    def update_has_data(self):
        """
        By default, assume that the gramplet has data.
        """
        self.set_has_data(True)

    def set_orientation(self, orientation):
        """
        Called when the gramplet orientation changes. A gramplet may override
        this if it has a preferred horizontal and vertical layout.

        :param orientation: A Gtk.Orientation (VERTCIAL or HORIZONTAL)
        :type orientation: int
        """



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2025       Steve Youngs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gi.repository import GObject

"""
Package providing filtering framework for Gramps.
"""

# -------------------------------------------------------------------------
#
# GTK
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from ..utils import no_match_primary_mask

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")


# -------------------------------------------------------------------------
#
# SearchBar
#
# -------------------------------------------------------------------------
class SearchBar:
    def __init__(self, dbstate, uistate, on_apply, apply_done=None, apply_clear=None):
        self.on_apply_callback = on_apply
        self.apply_done_callback = apply_done
        self.apply_clear_callback = apply_clear
        self.dbstate = dbstate
        self.uistate = uistate
        self.apply_text = ""
        self.visible = False

        self.searchbar = Gtk.Box()
        self.search_text = Gtk.Entry()
        self.find_button = Gtk.Button.new_with_mnemonic(_("_Find"))
        self.clear_button = Gtk.Button.new_with_mnemonic(_("_Clear"))
        self.search_list = Gtk.ComboBox()
        self.search_model = Gtk.ListStore(
            GObject.TYPE_STRING,  # rule name
            GObject.TYPE_INT,  # index of column to search
            GObject.TYPE_BOOLEAN,  # inversion: invert the search result if True
            GObject.TYPE_BOOLEAN,  # type : True = filter, False = search
            GObject.TYPE_PYOBJECT,  # filter object
        )

    def destroy(self):
        """Unset all things that can block garbage collection."""
        self.on_apply_callback = None
        self.apply_done_callback = None
        self.dbstate = None
        self.uistate = None

    def build(self):
        self.searchbar.set_spacing(4)
        self.search_list.connect("changed", self.search_changed)

        self.search_text.connect("key-press-event", self.key_press)
        self.search_text.connect("changed", self.text_changed)

        self.find_button.connect("clicked", self.apply_find_clicked)
        self.find_button.set_sensitive(False)

        self.clear_button.connect("clicked", self.apply_clear)
        self.clear_button.set_sensitive(False)

        self.searchbar.pack_start(self.search_list, False, True, 0)
        self.searchbar.pack_start(self.search_text, True, True, 0)
        self.searchbar.pack_end(self.clear_button, False, True, 0)
        self.searchbar.pack_end(self.find_button, False, True, 0)

        return self.searchbar

    def setup_searches(self, column_data):
        """
        column_data is a list of tuples:
        [(trans_col_name, index, use_exact), ...]
        """
        self.search_model.clear()
        old_value = self.search_list.get_active()

        cell = Gtk.CellRendererText()
        self.search_list.clear()
        self.search_list.pack_start(cell, True)
        self.search_list.add_attribute(cell, "text", 0)

        maxval = 0
        for col, index, exact in column_data:
            if exact:
                rule = _("%s is") % col
            else:
                rule = _("%s contains") % col
            self.search_model.append(row=[rule, index, False, False, None])
            maxval += 1
            if exact:
                rule = _("%s is not") % col
            else:
                rule = _("%s does not contain") % col
            self.search_model.append(row=[rule, index, True, False, None])
            maxval += 1

        self.search_list.set_model(self.search_model)
        if old_value == -1 or old_value >= maxval:
            self.search_list.set_active(0)
        else:
            self.search_list.set_active(old_value)

    def append_filter(self, filter_name, filter):
        self.search_model.append(row=[filter_name, -1, False, True, filter])

    def search_changed(self, obj):
        self.find_button.set_sensitive(True)
        # only make the search_text and clear_button widgets sensitive for searches
        node = self.search_list.get_active_iter()
        type = self.search_model.get_value(node, 3)
        self.clear_button.set_sensitive(not type)
        self.search_text.set_sensitive(not type)

    def text_changed(self, obj):
        text = obj.get_text()
        if self.apply_text == "" and text == "":
            self.find_button.set_sensitive(False)
            self.clear_button.set_sensitive(False)
        elif self.apply_text == text:
            self.find_button.set_sensitive(False)
            self.clear_button.set_sensitive(True)
        else:
            self.find_button.set_sensitive(True)
            self.clear_button.set_sensitive(True)

    def key_press(self, obj, event):
        if no_match_primary_mask(event.get_state()):
            if event.keyval in (_RETURN, _KP_ENTER):
                self.find_button.set_sensitive(False)
                self.clear_button.set_sensitive(True)
                self.apply_search()
        return False

    def apply_find_clicked(self, obj):
        self.apply_search()

    def apply_clear(self, obj):
        self.search_text.set_text("")
        self.apply_search()
        if self.apply_clear_callback is not None:
            self.apply_clear_callback()

    def get_value(self):
        text = str(self.search_text.get_text()).strip()
        node = self.search_list.get_active_iter()
        index = self.search_model.get_value(node, 1)
        inv = self.search_model.get_value(node, 2)
        type = self.search_model.get_value(node, 3)
        filter = self.search_model.get_value(node, 4)
        return (type, filter, False) if type else (type, (index, text, inv), False)

    def apply_search(self, current_model=None):
        self.apply_text = str(self.search_text.get_text())
        self.find_button.set_sensitive(False)
        self.uistate.status_text(_("Updating display..."))
        self.on_apply_callback()
        self.search_text.grab_focus()
        self.uistate.modify_statusbar(self.dbstate)

    def show(self):
        self.searchbar.show()
        self.visible = True

    def hide(self):
        self.searchbar.hide()
        self.visible = False

    def is_visible(self):
        return self.visible



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007       Brian G. Matherly
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Proxy class for the Gramps databases. Filter out all data marked private.
"""

# -------------------------------------------------------------------------
#
# Python libraries
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import logging

LOG = logging.getLogger(".citation")

# -------------------------------------------------------------------------
#
# Gramps libraries
#
# -------------------------------------------------------------------------
from ..lib import (
    MediaRef,
    Attribute,
    Address,
    EventRef,
    Person,
    PersonRef,
    Name,
    Source,
    RepoRef,
    Media,
    Place,
    Event,
    Family,
    ChildRef,
    Repository,
    LdsOrd,
    Surname,
    Citation,
    SrcAttribute,
    Note,
    Tag,
)
from .proxybase import ProxyDbBase


class PrivateProxyDb(ProxyDbBase):
    """
    A proxy to a Gramps database. This proxy will act like a Gramps database,
    but all data marked private will be hidden from the user.
    """

    def __init__(self, db):
        """
        Create a new PrivateProxyDb instance.
        """
        ProxyDbBase.__init__(self, db)

    def get_person_from_handle(self, handle):
        """
        Finds a Person in the database from the passed Gramps ID.
        If no such Person exists, None is returned.
        """
        person = self.db.get_person_from_handle(handle)
        if person and not person.get_privacy():
            return sanitize_person(self.db, person)
        return None

    def get_source_from_handle(self, handle):
        """
        Finds a Source in the database from the passed Gramps ID.
        If no such Source exists, None is returned.
        """
        source = self.db.get_source_from_handle(handle)
        if source and not source.get_privacy():
            return sanitize_source(self.db, source)
        return None

    def get_citation_from_handle(self, handle):
        """
        Finds a Citation in the database from the passed Gramps ID.
        If no such Citation exists, None is returned.
        """
        citation = self.db.get_citation_from_handle(handle)
        if citation and not citation.get_privacy():
            return sanitize_citation(self.db, citation)
        return None

    def get_media_from_handle(self, handle):
        """
        Finds an Object in the database from the passed Gramps ID.
        If no such Object exists, None is returned.
        """
        media = self.db.get_media_from_handle(handle)
        if media and not media.get_privacy():
            return sanitize_media(self.db, media)
        return None

    def get_place_from_handle(self, handle):
        """
        Finds a Place in the database from the passed Gramps ID.
        If no such Place exists, None is returned.
        """
        place = self.db.get_place_from_handle(handle)
        if place and not place.get_privacy():
            return sanitize_place(self.db, place)
        return None

    def get_event_from_handle(self, handle):
        """
        Finds a Event in the database from the passed Gramps ID.
        If no such Event exists, None is returned.
        """
        event = self.db.get_event_from_handle(handle)
        if event and not event.get_privacy():
            return sanitize_event(self.db, event)
        return None

    def get_family_from_handle(self, handle):
        """
        Finds a Family in the database from the passed Gramps ID.
        If no such Family exists, None is returned.
        """
        family = self.db.get_family_from_handle(handle)
        if family and not family.get_privacy():
            return sanitize_family(self.db, family)
        return None

    def get_repository_from_handle(self, handle):
        """
        Finds a Repository in the database from the passed Gramps ID.
        If no such Repository exists, None is returned.
        """
        repository = self.db.get_repository_from_handle(handle)
        if repository and not repository.get_privacy():
            return sanitize_repository(self.db, repository)
        return None

    def get_note_from_handle(self, handle):
        """
        Finds a Note in the database from the passed Gramps ID.
        If no such Note exists, None is returned.
        """
        note = self.db.get_note_from_handle(handle)
        if note and not note.get_privacy():
            # Nothing to sanitize in note object:
            return note
        return None

    def get_person_from_gramps_id(self, val):
        """
        Finds a Person in the database from the passed Gramps ID.
        If no such Person exists, None is returned.
        """
        person = self.db.get_person_from_gramps_id(val)
        if person and not person.get_privacy():
            return sanitize_person(self.db, person)
        return None

    def get_family_from_gramps_id(self, val):
        """
        Finds a Family in the database from the passed Gramps ID.
        If no such Family exists, None is returned.
        """
        family = self.db.get_family_from_gramps_id(val)
        if family and not family.get_privacy():
            return sanitize_family(self.db, family)
        return None

    def get_event_from_gramps_id(self, val):
        """
        Finds an Event in the database from the passed Gramps ID.
        If no such Event exists, None is returned.
        """
        event = self.db.get_event_from_gramps_id(val)
        if event and not event.get_privacy():
            return sanitize_event(self.db, event)
        return None

    def get_place_from_gramps_id(self, val):
        """
        Finds a Place in the database from the passed Gramps ID.
        If no such Place exists, None is returned.
        """
        place = self.db.get_place_from_gramps_id(val)
        if place and not place.get_privacy():
            return sanitize_place(self.db, place)
        return None

    def get_source_from_gramps_id(self, val):
        """
        Finds a Source in the database from the passed Gramps ID.
        If no such Source exists, None is returned.
        """
        source = self.db.get_source_from_gramps_id(val)
        if source and not source.get_privacy():
            return sanitize_source(self.db, source)
        return None

    def get_citation_from_gramps_id(self, val):
        """
        Finds a Citation in the database from the passed Gramps ID.
        If no such Citation exists, None is returned.
        """
        citation = self.db.get_citation_from_gramps_id(val)
        if citation and not citation.get_privacy():
            return sanitize_citation(self.db, citation)
        return None

    def get_media_from_gramps_id(self, val):
        """
        Finds a Media in the database from the passed Gramps ID.
        If no such Media exists, None is returned.
        """
        obj = self.db.get_media_from_gramps_id(val)
        if obj and not obj.get_privacy():
            return sanitize_media(self.db, obj)
        return None

    def get_repository_from_gramps_id(self, val):
        """
        Finds a Repository in the database from the passed Gramps ID.
        If no such Repository exists, None is returned.
        """
        repository = self.db.get_repository_from_gramps_id(val)
        if repository and not repository.get_privacy():
            return sanitize_repository(self.db, repository)
        return None

    def get_note_from_gramps_id(self, val):
        """
        Finds a Note in the database from the passed Gramps ID.
        If no such Note exists, None is returned.
        """
        note = self.db.get_note_from_gramps_id(val)
        if note and not note.get_privacy():
            return note
        return None

    # Define predicate functions for use by default iterator methods

    def include_person(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_person(handle)
        return obj and not obj.get_privacy()

    def include_family(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_family(handle)
        return obj and not obj.get_privacy()

    def include_event(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_event(handle)
        return obj and not obj.get_privacy()

    def include_source(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_source(handle)
        return obj and not obj.get_privacy()

    def include_citation(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_citation(handle)
        return obj and not obj.get_privacy()

    def include_place(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_place(handle)
        return obj and not obj.get_privacy()

    def include_media(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_media(handle)
        return obj and not obj.get_privacy()

    def include_repository(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_repository(handle)
        return obj and not obj.get_privacy()

    def include_note(self, handle):
        """
        Predicate returning True if object is to be included, else False
        """
        obj = self.get_unfiltered_note(handle)
        return obj and not obj.get_privacy()

    def get_default_person(self):
        """returns the default Person of the database"""
        person = self.db.get_default_person()
        if person and not person.get_privacy():
            return sanitize_person(self.db, person)
        return None

    def get_default_handle(self):
        """returns the default Person of the database"""
        handle = self.db.get_default_handle()
        if handle:
            person = self.db.get_person_from_handle(handle)
            if person and not person.get_privacy():
                return handle
        return None

    def has_person_handle(self, handle):
        """
        returns True if the handle exists in the current Person database.
        """
        person = self.db.get_person_from_handle(handle)
        if person and not person.get_privacy():
            return True
        return False

    def has_event_handle(self, handle):
        """
        returns True if the handle exists in the current Event database.
        """
        event = self.db.get_event_from_handle(handle)
        if event and not event.get_privacy():
            return True
        return False

    def has_source_handle(self, handle):
        """
        returns True if the handle exists in the current Source database.
        """
        source = self.db.get_source_from_handle(handle)
        if source and not source.get_privacy():
            return True
        return False

    def has_citation_handle(self, handle):
        """
        returns True if the handle exists in the current Citation database.
        """
        citation = self.db.get_citation_from_handle(handle)
        if citation and not citation.get_privacy():
            return True
        return False

    def has_place_handle(self, handle):
        """
        returns True if the handle exists in the current Place database.
        """
        place = self.db.get_place_from_handle(handle)
        if place and not place.get_privacy():
            return True
        return False

    def has_family_handle(self, handle):
        """
        Return True if the handle exists in the current Family database.
        """
        family = self.db.get_family_from_handle(handle)
        if family and not family.get_privacy():
            return True
        return False

    def has_object_handle(self, handle):
        """
        Return True if the handle exists in the current Mediadatabase.
        """
        object = self.db.get_media_from_handle(handle)
        if object and not object.get_privacy():
            return True
        return False

    def has_repository_handle(self, handle):
        """
        Return True if the handle exists in the current Repository database.
        """
        repository = self.db.get_repository_from_handle(handle)
        if repository and not repository.get_privacy():
            return True
        return False

    def has_note_handle(self, handle):
        """
        Return True if the handle exists in the current Note database.
        """
        note = self.db.get_note_from_handle(handle)
        if note and not note.get_privacy():
            return True
        return False

    def find_backlink_handles(self, handle, include_classes=None):
        """
        Find all objects that hold a reference to the object handle.
        Returns an iterator over a list of (class_name, handle) tuples.

        :param handle: handle of the object to search for.
        :type handle: database handle
        :param include_classes: list of class names to include in the results.
                                Default: None means include all classes.
        :type include_classes: list of class names

        This default implementation does a sequential scan through all
        the primary object databases and is very slow. Backends can
        override this method to provide much faster implementations that
        make use of additional capabilities of the backend.

        Note that this is a generator function, it returns a iterator for
        use in loops. If you want a list of the results use::

        >    result_list = list(find_backlink_handles(handle))
        """

        # This isn't done yet because it doesn't check if references are
        # private (like a MediaRef). It only checks if the
        # referenced object is private.

        objects = {
            "Person": self.db.get_person_from_handle,
            "Family": self.db.get_family_from_handle,
            "Event": self.db.get_event_from_handle,
            "Source": self.db.get_source_from_handle,
            "Citation": self.db.get_citation_from_handle,
            "Place": self.db.get_place_from_handle,
            "Media": self.db.get_media_from_handle,
            "Note": self.db.get_note_from_handle,
            "Repository": self.db.get_repository_from_handle,
        }

        handle_itr = self.db.find_backlink_handles(handle, include_classes)
        for class_name, handle in handle_itr:
            if class_name in objects:
                obj = objects[class_name](handle)
                if obj and not obj.get_privacy():
                    yield (class_name, handle)
            else:
                raise NotImplementedError
        return


def copy_media_ref_list(db, original_obj, clean_obj):
    """
    Copies media references from one object to another - excluding private
    references and references to private objects.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have private references
    :type original_obj: MediaBase
    :param clean_obj: Object that will have only non-private references
    :type original_obj: MediaBase
    :returns: Nothing
    """
    for media_ref in original_obj.get_media_list():
        if media_ref and not media_ref.get_privacy():
            handle = media_ref.get_reference_handle()
            media = db.get_media_from_handle(handle)
            if media and not media.get_privacy():
                clean_obj.add_media_reference(sanitize_media_ref(db, media_ref))


def copy_citation_ref_list(db, original_obj, clean_obj):
    """
    Copies citation references from one object to another - excluding references
    to private citations, and references to citations that refer to private
    sources.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have private references
    :type original_obj: CitationBase
    :param clean_obj: Object that will have only non-private references
    :type original_obj: CitationBase
    :returns: Nothing
    """
    for citation_handle in original_obj.get_citation_list():
        citation = db.get_citation_from_handle(citation_handle)
        if citation and not citation.get_privacy():
            handle = citation.get_reference_handle()
            source = db.get_source_from_handle(handle)
            if source and not source.get_privacy():
                clean_obj.add_citation(citation_handle)


def copy_notes(db, original_obj, clean_obj):
    """
    Copies notes from one object to another - excluding references to private
    notes.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have private references
    :type original_obj: NoteBase
    :param clean_obj: Object that will have only non-private references
    :type original_obj: NoteBase
    :returns: Nothing
    """
    for note_handle in original_obj.get_note_list():
        note = db.get_note_from_handle(note_handle)
        if note and not note.get_privacy():
            clean_obj.add_note(note_handle)


def copy_associations(db, original_obj, clean_obj):
    """
    Copies associations from one object to another - excluding
    references to private notes.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have private references
    :type original_obj: Base
    :param clean_obj: Object that will have only non-private references
    :type original_obj: Base
    :returns: Nothing
    """
    new_person_ref_list = []
    for person_ref in original_obj.get_person_ref_list():
        if person_ref and not person_ref.get_privacy():
            associated_person = db.get_person_from_handle(person_ref.ref)
            if associated_person and not associated_person.get_privacy():
                new_person_ref = sanitize_person_ref(db, person_ref)
                new_person_ref_list.append(new_person_ref)
    clean_obj.set_person_ref_list(new_person_ref_list)


def copy_attributes(db, original_obj, clean_obj):
    """
    Copies attributes from one object to another - excluding references to
    private attributes.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have private references
    :type original_obj: AttributeBase
    :param clean_obj: Object that will have only non-private references
    :type original_obj: AttributeBase
    :returns: Nothing
    """
    for attribute in original_obj.get_attribute_list():
        if attribute and not attribute.get_privacy():
            new_attribute = Attribute()
            new_attribute.set_type(attribute.get_type())
            new_attribute.set_value(attribute.get_value())
            copy_notes(db, attribute, new_attribute)
            copy_citation_ref_list(db, attribute, new_attribute)
            clean_obj.add_attribute(new_attribute)


def copy_srcattributes(db, original_obj, clean_obj):
    """
    Copies srcattributes from one object to another - excluding references to
    private srcattributes.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have private references
    :type original_obj: SrcAttributeBase
    :param clean_obj: Object that will have only non-private references
    :type original_obj: SrcAttributeBase
    :returns: Nothing
    """
    for attribute in original_obj.get_attribute_list():
        if attribute and not attribute.get_privacy():
            new_attribute = SrcAttribute()
            new_attribute.set_type(attribute.get_type())
            new_attribute.set_value(attribute.get_value())
            clean_obj.add_attribute(new_attribute)


def copy_urls(db, original_obj, clean_obj):
    """
    Copies urls from one object to another - excluding references to
    private urls.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have urls
    :type original_obj: UrlBase
    :param clean_obj: Object that will have only non-private urls
    :type original_obj: UrlBase
    :returns: Nothing
    """
    for url in original_obj.get_url_list():
        if url and not url.get_privacy():
            clean_obj.add_url(url)


def copy_lds_ords(db, original_obj, clean_obj):
    """
    Copies LDS ORDs from one object to another - excluding references to
    private LDS ORDs.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have LDS ORDs
    :type original_obj: LdsOrdBase
    :param clean_obj: Object that will have only non-private LDS ORDs
    :type original_obj: LdsOrdBase
    :returns: Nothing
    """
    for lds_ord in original_obj.get_lds_ord_list():
        if lds_ord and not lds_ord.get_privacy():
            clean_obj.add_lds_ord(sanitize_lds_ord(db, lds_ord))


def copy_addresses(db, original_obj, clean_obj):
    """
    Copies addresses from one object to another - excluding references to
    private addresses.

    :param db: Gramps database to which the references belongs
    :type db: DbBase
    :param original_obj: Object that may have addresses
    :type original_obj: AddressBase
    :param clean_obj: Object that will have only non-private addresses
    :type original_obj: AddressBase
    :returns: Nothing
    """
    for address in original_obj.get_address_list():
        if address and not address.get_privacy():
            clean_obj.add_address(sanitize_address(db, address))


def sanitize_lds_ord(db, lds_ord):
    """
    Create a new LdsOrd instance based off the passed LdsOrd
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the LdsOrd object belongs
    :type db: DbBase
    :param name: source LdsOrd object that will be copied with
                 privacy records removed
    :type name: LdsOrd
    :returns: 'cleansed' LdsOrd object
    :rtype: LdsOrd
    """
    new_lds_ord = LdsOrd()
    new_lds_ord.set_type(lds_ord.get_type())
    new_lds_ord.set_status(lds_ord.get_status())
    new_lds_ord.set_temple(lds_ord.get_temple())

    fam_handle = lds_ord.get_family_handle()
    if fam_handle:
        fam = db.get_family_from_handle(fam_handle)
        if fam and not fam.get_privacy():
            new_lds_ord.set_family_handle(fam_handle)

    new_lds_ord.set_date_object(lds_ord.get_date_object())

    place_handle = lds_ord.get_place_handle()
    if place_handle:
        place = db.get_place_from_handle(place_handle)
        if place and not place.get_privacy():
            new_lds_ord.set_place_handle(place_handle)

    copy_citation_ref_list(db, lds_ord, new_lds_ord)
    copy_notes(db, lds_ord, new_lds_ord)

    return new_lds_ord


def sanitize_address(db, address):
    """
    Create a new Address instance based off the passed Address
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param name: source Address object that will be copied with
                 privacy records removed
    :type name: Address
    :returns: 'cleansed' Address object
    :rtype: Address
    """
    new_address = Address()

    new_address.set_street(address.get_street())
    new_address.set_locality(address.get_locality())
    new_address.set_city(address.get_city())
    new_address.set_county(address.get_county())
    new_address.set_state(address.get_state())
    new_address.set_country(address.get_country())
    new_address.set_postal_code(address.get_postal_code())
    new_address.set_phone(address.get_phone())

    new_address.set_date_object(address.get_date_object())
    copy_citation_ref_list(db, address, new_address)
    copy_notes(db, address, new_address)

    return new_address


def sanitize_name(db, name):
    """
    Create a new Name instance based off the passed Name
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param name: source Name object that will be copied with
                 privacy records removed
    :type name: Name
    :returns: 'cleansed' Name object
    :rtype: Name
    """
    new_name = Name()
    new_name.set_group_as(name.get_group_as())
    new_name.set_sort_as(name.get_sort_as())
    new_name.set_display_as(name.get_display_as())
    new_name.set_call_name(name.get_call_name())
    new_name.set_nick_name(name.get_nick_name())
    new_name.set_family_nick_name(name.get_family_nick_name())
    new_name.set_type(name.get_type())
    new_name.set_first_name(name.get_first_name())
    new_name.set_suffix(name.get_suffix())
    new_name.set_title(name.get_title())
    new_name.set_date_object(name.get_date_object())
    new_name.set_surname_list(name.get_surname_list())

    copy_citation_ref_list(db, name, new_name)
    copy_notes(db, name, new_name)

    return new_name


def sanitize_media_ref(db, media_ref):
    """
    Create a new MediaRef instance based off the passed MediaRef
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the MediaRef object belongs
    :type db: DbBase
    :param source_ref: source MediaRef object that will be copied with
                       privacy records removed
    :type source_ref: MediaRef
    :returns: 'cleansed' MediaRef object
    :rtype: MediaRef
    """
    new_ref = MediaRef()
    new_ref.set_rectangle(media_ref.get_rectangle())

    new_ref.set_reference_handle(media_ref.get_reference_handle())
    copy_notes(db, media_ref, new_ref)
    copy_attributes(db, media_ref, new_ref)
    copy_citation_ref_list(db, media_ref, new_ref)

    return new_ref


def sanitize_citation(db, citation):
    """
    Create a new Citation instance based off the passed Citation
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param citation: source Citation object that will be copied with
                     privacy records removed
    :type citation: Citation
    :returns: 'cleansed' Citation object
    :rtype: Citation
    """
    new_citation = Citation()
    new_citation.set_date_object(citation.get_date_object())
    new_citation.set_page(citation.get_page())
    new_citation.set_confidence_level(citation.get_confidence_level())
    new_citation.set_reference_handle(citation.get_reference_handle())
    new_citation.set_gramps_id(citation.get_gramps_id())
    new_citation.set_handle(citation.get_handle())
    new_citation.set_change_time(citation.get_change_time())
    new_citation.set_tag_list(citation.get_tag_list())
    copy_srcattributes(db, citation, new_citation)
    copy_notes(db, citation, new_citation)
    copy_media_ref_list(db, citation, new_citation)

    return new_citation


def sanitize_event_ref(db, event_ref):
    """
    Create a new EventRef instance based off the passed EventRef
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param event_ref: source EventRef object that will be copied with
                      privacy records removed
    :type event_ref: EventRef
    :returns: 'cleansed' EventRef object
    :rtype: EventRef
    """
    new_ref = EventRef()

    new_ref.set_reference_handle(event_ref.get_reference_handle())
    new_ref.set_role(event_ref.get_role())
    copy_citation_ref_list(db, event_ref, new_ref)
    copy_notes(db, event_ref, new_ref)
    copy_attributes(db, event_ref, new_ref)

    return new_ref


def sanitize_person(db, person):
    """
    Create a new Person instance based off the passed Person
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param person: source Person object that will be copied with
                   privacy records removed
    :type person: Person
    :returns: 'cleansed' Person object
    :rtype: Person
    """
    new_person = Person()

    # copy gender
    new_person.set_gender(person.get_gender())
    new_person.set_gramps_id(person.get_gramps_id())
    new_person.set_handle(person.get_handle())
    new_person.set_change_time(person.get_change_time())
    new_person.set_tag_list(person.get_tag_list())

    # copy names if not private
    name = person.get_primary_name()
    if (name and name.get_privacy()) or (person and person.get_privacy()):
        # Do this so a person always has a primary name of some sort.
        name = Name()
        surn = Surname()
        surn.set_surname(_("Private"))
        name.set_surname_list([surn])
        name.set_primary_surname()
    else:
        name = sanitize_name(db, name)
    new_person.set_primary_name(name)

    # copy Family reference list
    for handle in person.get_family_handle_list():
        family = db.get_family_from_handle(handle)
        if family and not family.get_privacy():
            new_person.add_family_handle(handle)

    # copy Family reference list
    for handle in person.get_parent_family_handle_list():
        family = db.get_family_from_handle(handle)
        if not family:
            continue
        elif family.get_privacy():
            continue
        child_ref_list = family.get_child_ref_list()
        for child_ref in child_ref_list:
            if child_ref.get_reference_handle() == person.get_handle():
                if child_ref and not child_ref.get_privacy():
                    new_person.add_parent_family_handle(handle)
                break

    for name in person.get_alternate_names():
        if name and not name.get_privacy():
            new_person.add_alternate_name(sanitize_name(db, name))

    # copy event list
    for event_ref in person.get_event_ref_list():
        if event_ref and not event_ref.get_privacy():
            event = db.get_event_from_handle(event_ref.ref)
            if event and not event.get_privacy():
                new_person.add_event_ref(sanitize_event_ref(db, event_ref))

    # Copy birth and death after event list to maintain the order.
    # copy birth event
    event_ref = person.get_birth_ref()
    if event_ref and not event_ref.get_privacy():
        event = db.get_event_from_handle(event_ref.ref)
        if event and not event.get_privacy():
            new_person.set_birth_ref(sanitize_event_ref(db, event_ref))

    # copy death event
    event_ref = person.get_death_ref()
    if event_ref and not event_ref.get_privacy():
        event = db.get_event_from_handle(event_ref.ref)
        if event and not event.get_privacy():
            new_person.set_death_ref(sanitize_event_ref(db, event_ref))

    copy_addresses(db, person, new_person)
    copy_attributes(db, person, new_person)
    copy_citation_ref_list(db, person, new_person)
    copy_urls(db, person, new_person)
    copy_media_ref_list(db, person, new_person)
    copy_lds_ords(db, person, new_person)
    copy_notes(db, person, new_person)
    copy_associations(db, person, new_person)

    return new_person


def sanitize_source(db, source):
    """
    Create a new Source instance based off the passed Source
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param source: source Source object that will be copied with
                   privacy records removed
    :type source: Source
    :returns: 'cleansed' Source object
    :rtype: Source
    """
    new_source = Source()

    new_source.set_author(source.get_author())
    new_source.set_title(source.get_title())
    new_source.set_publication_info(source.get_publication_info())
    new_source.set_abbreviation(source.get_abbreviation())
    new_source.set_gramps_id(source.get_gramps_id())
    new_source.set_handle(source.get_handle())
    new_source.set_change_time(source.get_change_time())
    new_source.set_tag_list(source.get_tag_list())

    for repo_ref in source.get_reporef_list():
        if repo_ref and not repo_ref.get_privacy():
            handle = repo_ref.get_reference_handle()
            repo = db.get_repository_from_handle(handle)
            if repo and not repo.get_privacy():
                new_source.add_repo_reference(RepoRef(repo_ref))

    copy_srcattributes(db, source, new_source)
    copy_media_ref_list(db, source, new_source)
    copy_notes(db, source, new_source)

    return new_source


def sanitize_media(db, media):
    """
    Create a new Media instance based off the passed Media
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param media: source Media object that will be copied with
                  privacy records removed
    :type media: Media
    :returns: 'cleansed' Media object
    :rtype: Media
    """
    new_media = Media()

    new_media.set_mime_type(media.get_mime_type())
    new_media.set_path(media.get_path())
    new_media.set_description(media.get_description())
    new_media.set_gramps_id(media.get_gramps_id())
    new_media.set_handle(media.get_handle())
    new_media.set_change_time(media.get_change_time())
    new_media.set_date_object(media.get_date_object())
    new_media.set_tag_list(media.get_tag_list())
    new_media.set_checksum(media.get_checksum())

    copy_citation_ref_list(db, media, new_media)
    copy_attributes(db, media, new_media)
    copy_notes(db, media, new_media)

    return new_media


def sanitize_place(db, place):
    """
    Create a new Place instance based off the passed Place
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param place: source Place object that will be copied with
                  privacy records removed
    :type place: Place
    :returns: 'cleansed' Place object
    :rtype: Place
    """
    new_place = Place()

    new_place.set_title(place.get_title())
    new_place.set_gramps_id(place.get_gramps_id())
    new_place.set_handle(place.get_handle())
    new_place.set_change_time(place.get_change_time())
    new_place.set_longitude(place.get_longitude())
    new_place.set_latitude(place.get_latitude())
    new_place.set_alternate_locations(place.get_alternate_locations())
    new_place.set_name(place.get_name())
    new_place.set_alternative_names(place.get_alternative_names())
    new_place.set_type(place.get_type())
    new_place.set_code(place.get_code())
    new_place.set_placeref_list(place.get_placeref_list())
    new_place.set_tag_list(place.get_tag_list())

    copy_citation_ref_list(db, place, new_place)
    copy_notes(db, place, new_place)
    copy_media_ref_list(db, place, new_place)
    copy_urls(db, place, new_place)

    return new_place


def sanitize_event(db, event):
    """
    Create a new Event instance based off the passed Event
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param event: source Event object that will be copied with
                  privacy records removed
    :type event: Event
    :returns: 'cleansed' Event object
    :rtype: Event
    """
    new_event = Event()

    new_event.set_type(event.get_type())
    new_event.set_description(event.get_description())
    new_event.set_gramps_id(event.get_gramps_id())
    new_event.set_handle(event.get_handle())
    new_event.set_date_object(event.get_date_object())
    new_event.set_change_time(event.get_change_time())
    new_event.set_tag_list(event.get_tag_list())

    copy_citation_ref_list(db, event, new_event)
    copy_notes(db, event, new_event)
    copy_media_ref_list(db, event, new_event)
    copy_attributes(db, event, new_event)

    place_handle = event.get_place_handle()
    if place_handle:
        place = db.get_place_from_handle(place_handle)
        if place and not place.get_privacy():
            new_event.set_place_handle(place_handle)

    return new_event


def sanitize_family(db, family):
    """
    Create a new Family instance based off the passed Family
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param family: source Family object that will be copied with
                   privacy records removed
    :type family: Family
    :returns: 'cleansed' Family object
    :rtype: Family
    """
    new_family = Family()

    new_family.set_gramps_id(family.get_gramps_id())
    new_family.set_handle(family.get_handle())
    new_family.set_relationship(family.get_relationship())
    new_family.set_change_time(family.get_change_time())
    new_family.set_tag_list(family.get_tag_list())

    # Copy the father handle.
    father_handle = family.get_father_handle()
    if father_handle:
        father = db.get_person_from_handle(father_handle)
        if father and not father.get_privacy():
            new_family.set_father_handle(father_handle)

    # Copy the mother handle.
    mother_handle = family.get_mother_handle()
    if mother_handle:
        mother = db.get_person_from_handle(mother_handle)
        if mother and not mother.get_privacy():
            new_family.set_mother_handle(mother_handle)

    # Copy child references.
    for child_ref in family.get_child_ref_list():
        if child_ref and child_ref.get_privacy():
            continue
        child_handle = child_ref.get_reference_handle()
        child = db.get_person_from_handle(child_handle)
        if child and child.get_privacy():
            continue
        # Copy this reference
        new_ref = ChildRef()
        new_ref.set_reference_handle(child_ref.get_reference_handle())
        new_ref.set_father_relation(child_ref.get_father_relation())
        new_ref.set_mother_relation(child_ref.get_mother_relation())
        copy_notes(db, child_ref, new_ref)
        copy_citation_ref_list(db, child_ref, new_ref)
        new_family.add_child_ref(new_ref)

    # Copy event ref list.
    for event_ref in family.get_event_ref_list():
        if event_ref and not event_ref.get_privacy():
            event = db.get_event_from_handle(event_ref.ref)
            if event and not event.get_privacy():
                new_family.add_event_ref(sanitize_event_ref(db, event_ref))

    copy_citation_ref_list(db, family, new_family)
    copy_notes(db, family, new_family)
    copy_media_ref_list(db, family, new_family)
    copy_attributes(db, family, new_family)
    copy_lds_ords(db, family, new_family)

    return new_family


def sanitize_repository(db, repository):
    """
    Create a new Repository instance based off the passed Repository
    instance. The returned instance has all private records
    removed from it.

    :param db: Gramps database to which the Person object belongs
    :type db: DbBase
    :param repository: source Repository object that will be copied with
                       privacy records removed
    :type repository: Repository
    :returns: 'cleansed' Repository object
    :rtype: Repository
    """
    new_repository = Repository()

    new_repository.set_type(repository.get_type())
    new_repository.set_name(repository.get_name())
    new_repository.set_gramps_id(repository.get_gramps_id())
    new_repository.set_handle(repository.get_handle())
    new_repository.set_change_time(repository.get_change_time())
    new_repository.set_tag_list(repository.get_tag_list())

    copy_notes(db, repository, new_repository)
    copy_addresses(db, repository, new_repository)
    copy_urls(db, repository, new_repository)

    return new_repository


def sanitize_person_ref(db, person_ref: PersonRef) -> PersonRef:
    """
    Given a PersonRef, sanitize it.
    """
    new_person_ref = PersonRef()
    new_person_ref.ref = person_ref.ref
    new_person_ref.rel = person_ref.rel
    new_person_ref.private = person_ref.private
    copy_citation_ref_list(db, person_ref, new_person_ref)
    copy_notes(db, person_ref, new_person_ref)
    return new_person_ref



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2015       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .. import Rule
from ....utils.location import located_in

# -------------------------------------------------------------------------
#
# Typing modules
#
# -------------------------------------------------------------------------
from ....lib import Place
from ....db import Database


# -------------------------------------------------------------------------
#
# IsEnclosedBy
#
# -------------------------------------------------------------------------
class IsEnclosedBy(Rule):
    """
    Rule that checks for a place enclosed by another place
    """

    labels = [_("ID:"), _("Inclusive:")]
    name = _("Places enclosed by another place")
    description = _("Matches a place enclosed by a particular place")
    category = _("General filters")

    def prepare(self, db: Database, user):
        self.handle = None
        place = db.get_place_from_gramps_id(self.list[0])
        if place:
            self.handle = place.handle

    def apply_to_one(self, db: Database, place: Place) -> bool:
        if self.handle is None:
            return False
        if self.list[1] == "1" and place.handle == self.handle:
            return True
        if located_in(db, place.handle, self.handle):
            return True
        return False



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2004-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Finnish-specific classes for parsing and displaying dates.
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib.date import Date
from ._dateparser import DateParser
from ._datedisplay import DateDisplay
from ._datehandler import register_datehandler


# -------------------------------------------------------------------------
#
# Finnish parser
#
# This handles only dates where days and months are given as numeric, as:
# - That's how they are normally used in Finland
# - Parsing Finnish is much more complicated than English
# -------------------------------------------------------------------------
class DateParserFI(DateParser):
    # NOTE: these need to be in lower case because the "key" comparison
    # is done as lower case.  In the display method correct capitalization
    # can be used.

    modifier_to_int = {
        # examples:
        # - ennen 1.1.2005
        # - noin 1.1.2005
        "ennen": Date.MOD_BEFORE,
        "e.": Date.MOD_BEFORE,
        "noin": Date.MOD_ABOUT,
        "n.": Date.MOD_ABOUT,
        "alkaen": Date.MOD_FROM,
        "päättyen": Date.MOD_TO,
    }
    modifier_after_to_int = {
        # examples:
        # - 1.1.2005 jälkeen
        "jälkeen": Date.MOD_AFTER,
        "j.": Date.MOD_AFTER,
    }

    bce = ["ekr.", "ekr"]

    calendar_to_int = {
        "gregoriaaninen": Date.CAL_GREGORIAN,
        "greg.": Date.CAL_GREGORIAN,
        "juliaaninen": Date.CAL_JULIAN,
        "jul.": Date.CAL_JULIAN,
        "heprealainen": Date.CAL_HEBREW,
        "hepr.": Date.CAL_HEBREW,
        "islamilainen": Date.CAL_ISLAMIC,
        "isl.": Date.CAL_ISLAMIC,
        "ranskan vallankumouksen aikainen": Date.CAL_FRENCH,
        "ranskan v.": Date.CAL_FRENCH,
        "persialainen": Date.CAL_PERSIAN,
        "pers.": Date.CAL_PERSIAN,
        "svensk": Date.CAL_SWEDISH,
        "s": Date.CAL_SWEDISH,
    }

    quality_to_int = {
        "arviolta": Date.QUAL_ESTIMATED,
        "arv.": Date.QUAL_ESTIMATED,
        "laskettuna": Date.QUAL_CALCULATED,
        "lask.": Date.QUAL_CALCULATED,
    }

    def init_strings(self):
        DateParser.init_strings(self)
        self._text2 = re.compile(
            r"(\d+)?\.?\s+?%s\.?\s*((\d+)(/\d+)?)?\s*$" % self._mon_str, re.IGNORECASE
        )
        self._span = re.compile(r"(?P<start>.+)\s+(-)\s+(?P<stop>.+)", re.IGNORECASE)
        self._range = re.compile(
            r"(vuosien\s*)?(?P<start>.+)\s+ja\s+(?P<stop>.+)\s+välillä", re.IGNORECASE
        )


# -------------------------------------------------------------------------
#
# Finnish display
#
# -------------------------------------------------------------------------
class DateDisplayFI(DateDisplay):
    """
    Finnish language date display class.
    """

    _bce_str = "%s ekr."

    formats = (
        "VVVV-KK-PP (ISO)",
        "PP.KK.VVVV",
        "Päivä Kuukausi Vuosi",  # Day, full month name, year
    )
    # this definition must agree with its "_display_calendar" method

    display = DateDisplay.display_formatted

    def _display_calendar(self, date_val, long_months, short_months=None, inflect=""):
        # this must agree with its locale-specific "formats" definition

        if short_months is None:
            # Let the short formats work the same as long formats
            short_months = long_months

        if self.format == 0:
            return self.display_iso(date_val)
        elif self.format == 1:
            # numerical
            value = self.dd_dformat01(date_val)
        # elif self.format == 4:
        else:
            # day month_name year
            value = self.dd_dformat04(date_val, inflect, long_months)
        if date_val[2] < 0:
            # TODO fix BUG 7064: non-Gregorian calendars wrongly use BCE notation for negative dates
            return self._bce_str % value
        else:
            return value

    def dd_dformat04(self, date_val, inflect, long_months):
        """
        day month_name year -- for Finnish locale
        """
        year = self._slash_year(date_val[2], date_val[3])
        if date_val[0] == 0:
            if date_val[1] == 0:
                return year
            else:
                if inflect:
                    return self.format_long_month_year(
                        date_val[1], year, inflect, long_months
                    )
                else:
                    return "{long_month.forms[IN]} {year}".format(
                        long_month=long_months[date_val[1]], year=year
                    )
        else:
            if not hasattr(long_months[date_val[1]], "forms"):  # not a Lexeme
                return self.dd_dformat01(date_val)  # maybe the month is zero
            return "{day:d}. {long_month.forms[P]} {year}".format(
                day=date_val[0], long_month=long_months[date_val[1]], year=year
            )


# -------------------------------------------------------------------------
#
# Register classes
#
# -------------------------------------------------------------------------
register_datehandler(
    ("fi_FI", "fi", "finnish", "Finnish", ("%d.%m.%Y",)), DateParserFI, DateDisplayFI
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2021    Matthias Kemmer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""Heatmap web report."""


# ------------------------------------------------
# python modules
# ------------------------------------------------
import logging
from string import Template
from decimal import Decimal, getcontext


# ------------------------------------------------------------------------
#
# GRAMPS modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.utils.place import conv_lat_lon
from gramps.plugins.lib.libhtml import Html
from gramps.gen.display.name import displayer as _nd
from gramps.gen.display.place import displayer as _pd

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.gen.lib import (
    EventType,
    Event,
    Family,
    Name,
    Person,
)
from gramps.plugins.webreport.common import (
    alphabet_navigation,
    FULLCLEAR,
    AlphabeticIndex,
    MARKER_PATH,
    OPENLAYER,
    OSM_MARKERS,
    html_escape,
)
from gramps.plugins.webreport.multiselect import (
    get_surnames_list,
    get_tags_list,
)

_ = glocale.translation.sgettext
LOG = logging.getLogger(".Heatmap")
getcontext().prec = 8

try:
    _trans = glocale.get_addon_translator(__file__)
except ValueError:
    _trans = glocale.translation
_ = _trans.gettext


# ------------------------------------------------------------------------
#
# HeatmapPlaces Class
#
# ------------------------------------------------------------------------
class HeatmapPlace:
    """Class storing heatmap place data."""

    def __init__(self, name, latitude, longitude, count, titlename):
        self.name = name  # gramps_id
        self.lat = latitude
        self.lon = longitude
        self.count = count
        self.title_name = titlename
        self.events = []

    def to_list(self):
        return [self.lat, self.lon, self.count]

    def add_event(self, event_ID):
        self.events.append(event_ID)

    def get_all(self):
        return [self.name, self.lat, self.lon, self.count, self.title_name, self.events]


# ------------------------------------------------------------------------
#
# Heatmap Report Class
#
# ------------------------------------------------------------------------
class HeatmapPage(BasePage):
    """
    This class is responsible for displaying information about Heatmaps

    The base class 'BasePage' is initialised once for each page that is
    displayed.
    """

    def __init__(self, report, the_lang, the_title, the_filter):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title, the_filter)
        self.report = report
        self.sort_name = None
        self.person = None
        self.rel_class = None
        self.events = []
        self.places = []
        self.count = 0
        self.radius = int(report.options["radius"])
        self.blur = int(report.options["blur"])
        self.vallat = float(0.0)
        self.vallon = float(0.0)
        self.lookingfor = None
        self.selected_filter = the_filter
        self.display_pages(the_lang, the_title)

    def display_pages(self, the_lang, the_title):
        """
        Generate and output the pages under the Individuals tab, namely the
        individual index and the individual pages.

        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        default_types = [name[1] for name in EventType._DATAMAP]  # Initial local name
        custom_types = [name for name in self.r_db.get_event_types()]
        sorted_default_types = sorted([*default_types, *custom_types])
        selected = self.report.options["selected_evts"]

        def in_standard_event(value):
            result = False
            for evt in EventType._DATAMAP:
                if value == evt[1]:
                    result = True
            return result

        saved_evts = []
        for evt in selected:
            index = 0
            for typ in sorted_default_types:
                if index == int(evt) and not in_standard_event(typ):
                    self.events.append(typ)
                if index == int(evt):
                    # Save the local name
                    saved_evts.append(typ)
                index += 1
        default_types = [name for name in EventType._DATAMAP]  # Initial local name
        for evt in saved_evts:
            index = 0
            for typ in default_types:
                if typ[1] == evt:
                    self.events.append(typ[2])
                index += 1
        if "All Events" not in self.events:
            self.events.append("All Events")  # All events
        self.event_types = sorted([*default_types])

        self.heatmaplistpage(self.report, the_lang, the_title, self.events)
        progress_title = self.report.pgrs_title(the_lang)
        message = _("Creating heatmap pages for events")
        nbs = len(self.events)
        with self.r_user.progress(progress_title, message, nbs) as step:
            for event_type in sorted(self.events):
                step()
                self.heatmappage(
                    self.report, the_lang, the_title, event_type=event_type
                )
            step()

    def heatmaplistpage(self, report, the_lang, the_title, events):
        """Heatmap report class."""
        date = 0
        output_file, sio = self.report.create_file("heatmaps")
        self.uplink = False
        result = self.write_header(self._("Heatmaps"))
        indlistpage, dummy_head, dummy_body, outerwrapper = result
        self.cur_fname = self.report.cur_fname

        # begin events division
        with Html("div", class_="content", id="Events") as eventlist:
            outerwrapper += eventlist
            # Individual List page message
            msg = self._(
                "The next section contains an index of all the events for an Heatmap map."
            )
            eventlist += Html("p", msg, id="description")

            # Assemble the alphabeticIndex
            index = AlphabeticIndex(self.rlocale)
            for event in self.events:
                index.addRecord(self._(event), None)

            # Extract the buckets from the index
            index_list = []
            index.resetBucketIterator()
            while index.nextBucket():
                if index.bucketRecordCount != 0:
                    index_list.append(index.bucketLabel)
            # Output the navigation
            href_label = "event"
            alpha_nav = alphabet_navigation(
                index_list,
                self.rlocale,
                rtl=self.dir,
            )
            if alpha_nav:
                eventlist += alpha_nav

            # begin events table
            with Html(
                "table", class_="infolist primobjlist eventlist " + self.dir
            ) as table:
                eventlist += table

                # begin table head
                thead = Html("thead")
                table += thead

                trow = Html("tr")
                thead += trow

                trow.extend(
                    Html("th", label, class_=colclass, inline=True)
                    for (label, colclass) in [
                        [self._("Letter"), "ColumnLetter"],
                        [self._("Name", "Event Name"), "ColumnName"],
                        [self._("Description"), "ColumnDescription"],
                    ]
                )

                # begin table body
                tbody = Html("tbody")
                table += tbody

                # For each bucket, output the places in that bucket
                index.resetBucketIterator()
                output = []
                dup_index = 0
                while index.nextBucket():
                    if index.bucketRecordCount != 0:
                        bucket_letter = index.bucketLabel
                        bucket_link = bucket_letter
                        if bucket_letter in output:
                            bucket_link = "%s (%i)" % (bucket_letter, dup_index)
                            dup_index += 1
                        output.append(bucket_letter)
                        # Assemble all the events in this bucket into a dict for
                        # sorting
                        event_dict = {}
                        first_event = True
                        while index.nextRecord():
                            event_name = index.recordName
                            value = index.recordData
                            event_dict[event_name] = value

                            letter = bucket_letter
                            for event in events:
                                if self._(event) != event_name:
                                    continue
                                trow = Html("tr")
                                tbody += trow
                                tcell = Html("td", class_="ColumnLetter", inline=True)
                                trow += tcell
                                if first_event:
                                    first_event = False
                                    trow.attr = 'class = "BeginLetter"'
                                    ttle = (
                                        self._("Places beginning " "with letter %s")
                                        % letter
                                    )
                                    tcell += Html(
                                        "a",
                                        letter,
                                        name="%s_%s" % (href_label, letter),
                                        title=ttle,
                                        id_=bucket_link,
                                    )
                                else:
                                    tcell += "&nbsp;"
                                trow += Html(
                                    "td",
                                    self.heatmap_link(event),
                                    class_="ColumnName",
                                )
                                dname = self._(
                                    "Heatmap for %s within the selected filter"
                                ) % self._(event)
                                trow.extend(
                                    Html(
                                        "td",
                                        data or "&nbsp;",
                                        class_=colclass,
                                        inline=True,
                                    )
                                    for (colclass, data) in [
                                        ["ColumnDescription", dname],
                                    ]
                                )

        people = self.report.options["selected_surnames"]
        if people:
            # begin surnames division
            with Html("div", class_="content", id="Surnames") as surnamelist:
                outerwrapper += surnamelist
                msg = self._(
                    "The next section contains an index of all the surnames for an Heatmap map."
                )
                surnamelist += Html("p", msg, id="description")
                selected_people = get_surnames_list(self.r_db)
                index1 = AlphabeticIndex(self.rlocale)
                surnames = []
                for item in people:
                    index = 0
                    for value in sorted(
                        selected_people, key=lambda x: x[1], reverse=True
                    ):
                        if int(item) == index:
                            surnames.append(value[0])
                            index1.addRecord(value[0], None)
                            self.heatmappage(
                                self.report, the_lang, the_title, surname=value
                            )
                        index += 1

                # Extract the buckets from the index
                index_list = []
                index1.resetBucketIterator()
                while index1.nextBucket():
                    if index1.bucketRecordCount != 0:
                        index_list.append(index1.bucketLabel)
                # Output the navigation
                href_label = "surname"
                alpha_nav = alphabet_navigation(
                    index_list,
                    self.rlocale,
                    rtl=self.dir,
                )
                if alpha_nav:
                    surnamelist += alpha_nav

                # begin surnames table1
                with Html(
                    "table", class_="infolist primobjlist surnamelist " + self.dir
                ) as table1:
                    surnamelist += table1

                    # begin table1 head
                    thead = Html("thead")
                    table1 += thead

                    trow = Html("tr")
                    thead += trow

                    trow.extend(
                        Html("th", label, class_=colclass, inline=True)
                        for (label, colclass) in [
                            [self._("Letter"), "ColumnLetter"],
                            [self._("Surname", "Name"), "ColumnName"],
                            [self._("Description"), "ColumnDescription"],
                        ]
                    )

                    # begin table1 body
                    tbody = Html("tbody")
                    table1 += tbody

                    # For each bucket, output the places in that bucket
                    index1.resetBucketIterator()
                    output = []
                    dup_index = 0
                    while index1.nextBucket():
                        if index1.bucketRecordCount != 0:
                            bucket_letter = index1.bucketLabel
                            bucket_link = bucket_letter
                            if bucket_letter in output:
                                bucket_link = "%s (%i)" % (bucket_letter, dup_index)
                                dup_index += 1
                            output.append(bucket_letter)
                            # Assemble all the surnames in this bucket into a dict for
                            # sorting
                            surname_dict = {}
                            first_surname = True
                            while index1.nextRecord():
                                surname_name = index1.recordName
                                value = index1.recordData
                                surname_dict[surname_name] = value

                                letter = bucket_letter
                                for surname in surnames:
                                    if self._(surname) != surname_name:
                                        continue
                                    trow = Html("tr")
                                    tbody += trow
                                    tcell = Html(
                                        "td", class_="ColumnLetter", inline=True
                                    )
                                    trow += tcell
                                    if first_surname:
                                        first_surname = False
                                        trow.attr = 'class = "BeginLetter"'
                                        ttle = (
                                            self._("Places beginning " "with letter %s")
                                            % letter
                                        )
                                        tcell += Html(
                                            "a",
                                            letter,
                                            name="%s_%s" % (href_label, letter),
                                            title=ttle,
                                            id_=bucket_link,
                                        )
                                    else:
                                        tcell += "&nbsp;"
                                    trow += Html(
                                        "td",
                                        self.heatmap_link(surname),
                                        class_="ColumnName",
                                    )
                                    dname = self._(
                                        "Heatmap for %s within the selected filter"
                                    ) % self._(surname)
                                    trow.extend(
                                        Html(
                                            "td",
                                            data or "&nbsp;",
                                            class_=colclass,
                                            inline=True,
                                        )
                                        for (colclass, data) in [
                                            ["ColumnDescription", dname],
                                        ]
                                    )

        sel_tags = self.report.options["selected_tags"]
        if sel_tags:
            # begin tags division
            with Html("div", class_="content", id="Surnames") as taglist:
                outerwrapper += taglist
                msg = self._(
                    "The next section contains an index of all the tags for an Heatmap map."
                )
                taglist += Html("p", msg, id="description")
                selected_tags = get_tags_list(self.r_db)
                index2 = AlphabeticIndex(self.rlocale)
                tags = []
                for item in sel_tags:
                    index = 0
                    for value in selected_tags:
                        if int(item) == index:
                            tags.append(value[0])
                            index2.addRecord(value[0], None)
                            self.heatmappage(
                                self.report, the_lang, the_title, tag=value
                            )
                        index += 1

                # Extract the buckets from the index
                index_list = []
                index2.resetBucketIterator()
                while index2.nextBucket():
                    if index2.bucketRecordCount != 0:
                        index_list.append(index2.bucketLabel)
                # Output the navigation
                href_label = "tag"
                alpha_nav = alphabet_navigation(
                    index_list,
                    self.rlocale,
                    rtl=self.dir,
                )
                if alpha_nav:
                    taglist += alpha_nav

                # begin tags table1
                with Html(
                    "table", class_="infolist primobjlist taglist " + self.dir
                ) as table1:
                    taglist += table1

                    # begin table1 head
                    thead = Html("thead")
                    table1 += thead

                    trow = Html("tr")
                    thead += trow

                    trow.extend(
                        Html("th", label, class_=colclass, inline=True)
                        for (label, colclass) in [
                            [self._("Letter"), "ColumnLetter"],
                            [self._("Tag"), "ColumnName"],
                            [self._("Description"), "ColumnDescription"],
                        ]
                    )

                    # begin table1 body
                    tbody = Html("tbody")
                    table1 += tbody

                    # For each bucket, output the places in that bucket
                    index2.resetBucketIterator()
                    output = []
                    dup_index = 0
                    while index2.nextBucket():
                        if index2.bucketRecordCount != 0:
                            bucket_letter = index2.bucketLabel
                            bucket_link = bucket_letter
                            if bucket_letter in output:
                                bucket_link = "%s (%i)" % (bucket_letter, dup_index)
                                dup_index += 1
                            output.append(bucket_letter)
                            # Assemble all the tags in this bucket into a dict for
                            # sorting
                            tag_dict = {}
                            first_tag = True
                            while index2.nextRecord():
                                tag_name = index2.recordName
                                value = index2.recordData
                                tag_dict[tag_name] = value

                                letter = bucket_letter
                                for tag in tags:
                                    if self._(tag) != self._(tag_name):
                                        continue
                                    trow = Html("tr")
                                    tbody += trow
                                    tcell = Html(
                                        "td", class_="ColumnLetter", inline=True
                                    )
                                    trow += tcell
                                    if first_tag:
                                        first_tag = False
                                        trow.attr = 'class = "BeginLetter"'
                                        ttle = (
                                            self._("Places beginning " "with letter %s")
                                            % letter
                                        )
                                        tcell += Html(
                                            "a",
                                            letter,
                                            name="%s_%s" % (href_label, letter),
                                            title=ttle,
                                            id_=bucket_link,
                                        )
                                    else:
                                        tcell += "&nbsp;"
                                    trow += Html(
                                        "td",
                                        self.heatmap_link(tag),
                                        class_="ColumnName",
                                    )
                                    dname = self._(
                                        "Heatmap for %s within the selected filter"
                                    ) % self._(tag)
                                    trow.extend(
                                        Html(
                                            "td",
                                            data or "&nbsp;",
                                            class_=colclass,
                                            inline=True,
                                        )
                                        for (colclass, data) in [
                                            ["ColumnDescription", dname],
                                        ]
                                    )

        # create footer section
        footer = self.write_footer(date)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.report.cur_fname = self.cur_fname
        self.xhtml_writer(indlistpage, output_file, sio, date)

    def heatmappage(
        self, report, the_lang, the_title, event_type=None, surname=None, tag=None
    ):
        """Heatmap report class."""
        date = 0
        self.places = []
        self.tracelife = []
        self.count = 0
        self.vallat = float(0.0)
        self.vallon = float(0.0)
        self.uplink = True
        if event_type:
            self.lookingfor = self._(event_type)
            output_file, sio = self.report.create_file(
                event_type.replace(" ", ""), "heat"
            )
            result = self.write_header(event_type)
            events = self.get_all_events(event_type)
            msg = self._(
                "This section contains the map for all people with the event: {event_type} ({selected_filter})"
            ).format(event_type=event_type, selected_filter=self.selected_filter)
            if not len(self.places):
                msg += self._(" which has no event for the selected filter.")
            tracelife = self.create_tracelife(self.places, events)
        elif surname:
            selected = []
            output_file, sio = self.report.create_file(
                surname[0].replace(" ", ""), "heat"
            )
            result = self.write_header(surname[0])
            selected = self.get_surname_events(surname[2])
            msg = self._(
                "This section contains the map of all events for: {surname} ({selected_filter})"
            ).format(surname=surname[0], selected_filter=self.selected_filter)
            if not len(self.places):
                msg += self._(" which has no event for the selected filter.")
            tracelife = self.create_tracelife(self.places, selected)
        elif tag:
            output_file, sio = self.report.create_file(tag[0].replace(" ", ""), "heat")
            result = self.write_header(tag[0])
            selected = self.get_tag_events(tag)
            msg = self._(
                "This section contains the map for all people, events with the tag: {tag} ({selected_filter})"
            ).format(tag=tag[0], selected_filter=self.selected_filter)
            if not len(self.places):
                msg += self._(" which has no event for the selected filter.")
            tracelife = self.create_tracelife(self.places, tag)
        heatmappage, head, dummy_body, outerwrapper = result

        minx, maxx = Decimal("0.00000001"), Decimal("0.00000001")
        miny, maxy = Decimal("0.00000001"), Decimal("0.00000001")
        xwidth, yheight = [], []
        midx_, midy_, dummy_spanx, spany = [None] * 4

        # add narrative-maps CSS...
        if the_lang and not self.usecms:
            fname = "/".join(["..", "css", "narrative-maps.css"])
        else:
            fname = "/".join(["css", "narrative-maps.css"])
        url = self.report.build_url_fname(fname, None, self.uplink)
        head += Html(
            "link",
            href=url,
            type="text/css",
            media="screen",
            rel="stylesheet",
        )

        # add MapService specific javascript code. We use only OSM maps.
        src_js = "https://"
        src_js += "ajax.googleapis.com/ajax/libs/jquery/1.9.1/jquery.min.js"
        head += Html("script", type="text/javascript", src=src_js, inline=True)
        olv = self.report.options["ol_version"]
        build = "legacy"
        if olv < "v7.0.0":
            build = "build"
        if olv == "latest":
            build = "legacy"
        src_js = ("https://openlayers.org/en/" "%(ver)s/%(bld)s/ol.js") % {
            "ver": olv,
            "bld": build,
        }
        head += Html("script", type="text/javascript", src=src_js, inline=True)
        css = "legacy"
        if olv < "v7.0.0":
            css = "css"
        if olv == "latest":
            css = "legacy"
        url = ("https://openlayers.org/en/" "%(ver)s/%(css)s/ol.css") % {
            "ver": olv,
            "css": css,
        }
        head += Html("link", href=url, type="text/css", rel="stylesheet")

        src_js = ("https://openlayers.org/en/" "%(ver)s/%(bld)s/ol.js") % {
            "ver": olv,
            "bld": build,
        }
        head += Html("script", href=url, type="text/javascript", rel="stylesheet")

        # begin events division
        with Html("div", class_="content", id="events") as taglist:
            outerwrapper += taglist
            # Individual List page message
            taglist += Html("p", msg, id="description")

        # this is the style element where the Map is held in the CSS...
        with Html("div", id="map_canvas") as canvas:
            outerwrapper += canvas

        # add div for popups.
        with Html("div", id="popup", class_="ol-popup", inline=True) as popup:
            outerwrapper += popup
            popup += Html("a", href="#", id="popup-closer", class_="ol-popup-closer")
            popup += Html("div", id="popup-title", class_="ol-popup-title")
            popup += Html("div", id="popup-content", class_="ol-popup-content")
        with Html("div", id="tooltip", class_="ol-popup", inline=True) as tooltip:
            outerwrapper += tooltip
            tooltip += Html("div", id="tooltip-content")

        start_zoom = self.report.options["start_zoom"]

        # Collect heatmap coordinates
        heatmap_data = [place.to_list() for place in self.places]

        # Substitute the HTML template source code with heatmap report data
        if self.count:
            start_lat = self.vallat / self.count
            start_lon = self.vallon / self.count
        else:
            start_lat = 0.0
            start_lon = 0.0

        # begin javascript inline code...
        with Html(
            "script",
            deter="deter",
            style="width =100%; height =100%;",
            type="text/javascript",
            indent=False,
        ) as jsc:
            outerwrapper += jsc

            # Link to Gramps marker
            fname = "/".join(["images", "marker.png"])
            marker_path = self.report.build_url_image(
                "marker.png", "images", self.uplink
            )

            latitude = start_lat
            longitude = start_lon
            jsc += MARKER_PATH % marker_path
            # we are using OpenStreetMap
            if midy_ is None:
                jsc += OSM_MARKERS % (
                    "heatmap",
                    tracelife,
                    longitude,
                    latitude,
                    start_zoom,
                    self.radius,
                    self.blur,
                )
            else:
                jsc += OSM_MARKERS % (
                    "heatmap",
                    tracelife,
                    midy_,
                    midx_,
                    start_zoom,
                    self.radius,
                    self.blur,
                )
            jsc += OPENLAYER

        # create footer section
        footer = self.write_footer(date)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(heatmappage, output_file, sio, date)

    def get_all_events(self, event_type):
        """Get all relevant events from the current filter."""
        if self.lookingfor == self._("All Events"):
            for event_hdl in sorted(self.report.obj_dict[Event]):
                event = self.r_db.get_event_from_handle(event_hdl)
                self.select_event(event, evt=None)
        else:
            for index, name in enumerate(self.event_types):
                for event_hdl in sorted(self.report.obj_dict[Event]):
                    event = self.r_db.get_event_from_handle(event_hdl)
                    self.select_event(event, evt=event_type)

    def get_tag_events(self, tag):
        """Get all relevant tags."""
        selected = []
        for handle in self.report.obj_dict[Person]:
            person = self.r_db.get_person_from_handle(handle)
            if person:
                tags_list = person.get_tag_list()
                if tag[1] in tags_list:
                    selected.append((Person, handle))
                    self.get_place(person.get_event_ref_list())
        for handle in self.report.obj_dict[Family]:
            family = self.r_db.get_family_from_handle(handle)
            if family:
                tags_list = family.get_tag_list()
                if tag[1] in tags_list:
                    selected.append((Family, handle))
                    fam_evt_ref_list = family.get_event_ref_list()
                    if fam_evt_ref_list:
                        self.get_place(fam_evt_ref_list)
        for handle in self.report.obj_dict[Event]:
            event = self.r_db.get_event_from_handle(handle)
            if event:
                tags_list = event.get_tag_list()
                if tag[1] in tags_list:
                    selected.append((Event, handle))
                    self.get_place([event])
        return selected

    def get_surname_events(self, person_handle):
        selected_surname = []
        for handle in person_handle:
            if handle not in self.report.obj_dict[Person]:
                continue
            selected_surname.append(handle)
            events = self.get_events(handle)
            if events:
                self.get_place(events)
        return selected_surname

    def select_event(self, event, evt=None):
        if evt:
            if event.get_type().xml_str() == evt:
                handle = event.get_place_handle()
                if handle:
                    place = self.r_db.get_place_from_handle(handle)
                    self.check_place(place, event)
        else:
            handle = event.get_place_handle()
            if handle:
                place = self.r_db.get_place_from_handle(handle)
                self.check_place(place, event)

    def get_events(self, person_h, force=False):
        """Get all relevant events of a person."""
        person = self.r_db.get_person_from_handle(person_h)
        event_list = []
        for event_ref in person.get_event_ref_list():
            event = self.r_db.get_event_from_handle(event_ref.ref)
            event_list.append(event_ref)
        # select all events for the associated families
        fam_handle_list = person.get_family_handle_list()
        if fam_handle_list:
            for family_handle in fam_handle_list:
                family = self.r_db.get_family_from_handle(family_handle)
                if family:
                    fam_evt_ref_list = family.get_event_ref_list()
                    if fam_evt_ref_list:
                        for evt_ref in fam_evt_ref_list:
                            event_list.append(evt_ref)
        return event_list

    def get_place(self, events):
        """Get an event places and call check_place."""
        for event_ref in events:
            if not isinstance(event_ref, Event):
                event = self.r_db.get_event_from_handle(event_ref.ref)
            else:
                event = event_ref
            handle = event.get_place_handle()
            if handle:
                place = self.r_db.get_place_from_handle(handle)
                self.check_place(place, event)

    def check_place(self, place, event):
        """Check the place for latitude and longitude."""
        lat, lon = conv_lat_lon(place.get_latitude(), place.get_longitude(), "D.D8")
        name = place.get_gramps_id()
        placetitle = _pd.display(self.r_db, place, fmt=0)
        eid = event.get_gramps_id()

        if lat and lon:
            for placex in self.places:
                if name == placex.name:
                    placex.count += 1
                    if not (eid in placex.events):
                        placex.add_event(event.get_gramps_id())
                        self.vallat += float(lat)
                        self.vallon += float(lon)
                        self.count += 1
            found = None
            for placex in self.places:
                if name == placex.name:
                    found = placex
            if not found:
                self.places.append(HeatmapPlace(name, lat, lon, 1, placetitle))
                for placex in self.places:
                    if name == placex.name:
                        self.count += 1
                        placex.add_event(event.get_gramps_id())
                        self.vallat += float(lat)
                        self.vallon += float(lon)
        else:
            for place_ref in place.get_placeref_list():
                place_new = self.r_db.get_place_from_handle(place_ref.ref)
                self.check_place(place_new, event)

    def create_tracelife(self, places, handles):
        number_markers = len(places)
        places_list = []
        for place in places:
            places_list.append(place.get_all())
        sorted_places = sorted(places_list, key=lambda x: x[0])
        placetitle = "unknown"
        tracelife = "["
        if number_markers > 0:
            for index in range(0, number_markers):
                (name, latitude, longitude, count, placetitle, events) = sorted_places[
                    index
                ]
                debug = 0
                links = "''"
                for event_ID in events:
                    event = self.r_db.get_event_from_gramps_id(event_ID)
                    if event:
                        handle = event.get_place_handle()
                        if handle:
                            place = self.r_db.get_place_from_handle(handle)
                            placename = _pd.display(self.r_db, place, fmt=0)
                            p_fname = (
                                self.report.build_url_fname(handle, "plc", self.uplink)
                                + self.ext
                            )
                            placetitle = ' <a href="%s">%s</a>' % (p_fname, placename)

                        bkref_list = self.report.bkref_dict[Event][event.handle]
                        for ref in bkref_list:
                            (bkref_class, bkref_hdle, role) = ref
                            if bkref_class == Person:
                                person = self.r_db.get_person_from_handle(bkref_hdle)
                                links = self.__create_links_tracelife(
                                    links,
                                    person,
                                    placetitle,
                                    latitude,
                                    longitude,
                                    ref,
                                    event,
                                )
                debug = 0
                tracelife = self._create_family_tracelife(
                    tracelife, placetitle, latitude, longitude, links
                )
            tracelife += "]"
        else:
            tracelife += "]"
        return tracelife

    def _create_family_tracelife(
        self, tracelife, placetitle, latitude, longitude, links
    ):
        """
        creates individual family tracelife map events

        @param: tracelife  -- The family event list
        @param: placetitle -- The place title to add to the event list
        @param: latitude   -- The latitude for this place
        @param: longitude  -- The longitude for this place
        @param: links      -- Used to add links in the popup html page
        """
        tracelife += """[%f, %f, \'%s\', %s],""" % (
            float(longitude),
            float(latitude),
            placetitle.replace("'", "\\'"),
            links,
        )
        return tracelife

    def __create_links_tracelife(
        self, links, person, placetitle, latitude, longitude, ref, event
    ):
        """
        creates individual family events for a marker

        @param: links      -- Used to add links in the popup html page
        @param: person     -- family for this person
        @param: placetitle -- The place title to add to the event list
        @param: latitude   -- The latitude for this place
        @param: longitude  -- The longitude for this place
        @param: event      -- The event for which we are working
        @param: ref        -- The back reference and role for the event
        """
        if not person:
            return links
        url_fct = self.report.build_url_fname_html
        ppl_fct = self.r_db.get_person_from_handle
        ln_str = "<a href='%s' title='%s' target='_self'>%s</a>"
        ppl_lnk = ""
        (bkref_class, bkref_hdle, role) = ref
        if role == "Marriage" or role == "Divorce" or role == "Family":
            url = url_fct(bkref_hdle, "fam", self.uplink)
            fam_fct = self.r_db.get_family_from_handle
            fam = fam_fct(bkref_hdle)
            fam_name = self.report.get_family_name(fam)
            ppl_lnk = ln_str % (url, fam.get_gramps_id(), fam_name)
            if "<p>" in links:
                links += ' + "<br>%s"' % ppl_lnk
            else:
                links = '"<p>%s"' % ppl_lnk
        elif bkref_class == Person:
            pers = ppl_fct(bkref_hdle)
            url = url_fct(bkref_hdle, "ppl", self.uplink)
            ppl_lnk = ln_str % (
                url,
                pers.get_gramps_id(),
                self.get_name(pers).replace('"', ""),
            )
            evt_type = self._(event.get_type().xml_str())
            evt_date = self.rlocale.get_date(event.get_date_object())
            if self.report.options["inc_events"]:
                url = url_fct(event.get_handle(), "evt", self.uplink)
                evt_lnk = ln_str % (url, evt_date, evt_type)
            else:
                evt_lnk = evt_type
            if evt_date:
                evt_lnk += " (" + evt_date + ")"
            if "<p>" in links:
                links += ' + "<br>%s"' % (ppl_lnk + self._(":") + " " + evt_lnk)
            else:
                links = '"<p>%s"' % (ppl_lnk + self._(":") + " " + evt_lnk)
        return links
